import numpy as np


def AdjustByConstant(
    EM, AdjustVect, k=2, divMu=10, divLambda=10, verbose=False
):
    """
    Adjust elastic matrix by dividing the values of lambda and mu of stars

    GraphInfo the elpigraph structure updated
    k the largest orger of strars to leave unadjusted. e.g., if k = 2 only branching points will be adjusted
    divLambda the value used to divide the lambda coefficients
    divMu the value used to divide the mu coefficients

    @return


    @examples
    """
    EM2 = EM.copy()

    StarOrder = np.sum(EM > 0, axis=1) - 1
    StarOrder[StarOrder == 0] = 1

    ToAdjust = np.where((~np.array(AdjustVect)) & (StarOrder > k))[0]
    n_updated = len(ToAdjust)
    if verbose:
        print(n_updated, "values of the elastic matrix have been updated")

    if n_updated > 0:
        for i in ToAdjust:
            EM2[i, :] = EM[i, :] / divLambda
            EM2[:, i] = EM[:, i] / divLambda
            EM2[i, i] = EM[i, i] / divMu

            AdjustVect[i] = True
    return (EM2, AdjustVect)



import numpy as np
import elpigraph
from .src.PCA import PCA, TruncPCA, PCA_gpu, TruncSVD_gpu
from .src.distutils import PartialDistance


def computeElasticPrincipalCircle(
    X,
    NumNodes,
    NumEdges=float("inf"),
    InitNodes=3,
    Lambda=0.01,
    Mu=0.1,
    GrammarOptimization=False,
    MaxSteps=float("inf"),
    GrammarOrder=["Grow", "Shrink"],
    MaxNumberOfIterations=10,
    TrimmingRadius=float("inf"),
    eps=0.01,
    Do_PCA=True,
    InitNodePositions=None,
    AdjustVect=None,
    ElasticMatrix=None,
    InitEdges=None,
    CenterData=True,
    ComputeMSEP=True,
    verbose=False,
    ShowTimer=False,
    ReduceDimension=None,
    # drawAccuracyComplexity = True,
    # drawPCAView = True,
    # drawEnergy = True,
    n_cores=1,
    # ClusType = "Sock",
    MinParOp=20,
    nReps=1,
    # ParallelRep = False,
    Subsets=list(),
    ProbPoint=1,
    Mode=1,
    FinalEnergy="Base",
    alpha=0,
    beta=0,
    # gamma = 0,
    # FastSolve = False,
    Configuration="Line",
    ICOver=None,
    DensityRadius=None,
    AvoidSolitary=False,
    EmbPointProb=1,
    PointWeights=None,
    SampleIC=True,
    AvoidResampling=True,
    AdjustElasticMatrix=None,
    AdjustElasticMatrix_Initial=None,
    Lambda_Initial=None,
    Mu_Initial=None,
    DisplayWarnings=False,
    StoreGraphEvolution=False,
    GPU=False,
    FixNodesAtPoints=[],
    pseudotime=None,
    pseudotimeLambda=0.01,
    label=None,
    labelLambda=0.01,
    MaxNumberOfGraphCandidatesDict={
        "AddNode2Node": float("inf"),
        "BisectEdge": float("inf"),
        "RemoveNode": float("inf"),
        "ShrinkEdge": float("inf"),
    },
):

    """
    Construct a principal elastic circle

    This function is a wrapper to the computeElasticPrincipalGraph function that constructs the appropriate initial graph and grammars
    when constructing a circle

    X: numerical 2D matrix,
        the n-by-m matrix with the position of n m-dimensional points
    NumNodes: integer,
        the number of nodes of the principal graph
    Lambda: real,
        the lambda parameter used the compute the elastic energy
    Mu: real,
        the lambda parameter used the compute the elastic energy
    InitNodes: integer,
        number of points to include in the initial graph
    MaxNumberOfIterations: integer,
        maximum number of steps to embed the nodes in the data
    TrimmingRadius: real,
        maximal distance of point from a node to affect its embedment
    eps: real,
        minimal relative change in the position of the nodes to stop embedment
    Do_PCA: boolean,
        should data and initial node positions be PCA trnasformed?
    InitNodePositions: numerical 2D matrix,
        the k-by-m matrix with k m-dimensional positions of the nodes
    in the initial step
    InitEdges: numerical 2D matrix,
        the e-by-2 matrix with e end-points of the edges connecting the nodes
    ElasticMatrix: numerical 2D matrix,
        the k-by-k elastic matrix
    CenterData: boolean,
        should data and initial node positions be centered?
    ComputeMSEP: boolean,
        should MSEP be computed when building the report?
    verbose: boolean,
        should debugging information be reported?
    ShowTimer: boolean,
        should the time to construct the graph be computed and reported for each step?
    ReduceDimension: integer vector,
        vector of principal components to retain when performing
        dimensionality reduction. If None all the components will be used
    drawAccuracyComplexity: boolean,
        should the accuracy VS complexity plot be reported?
    drawPCAView: boolean,
        should a 2D plot of the points and pricipal curve be dranw for the final configuration?
    drawEnergy: boolean,
        should changes of evergy VS the number of nodes be reported?
    n:.cores
        either an integer (indicating the number of cores to used for the creation of a cluster) or
    cluster structure returned, e.g., by makeCluster. If a cluster structure is used, all the nodes must contains X
    (this is done using clusterExport)
    MinParOP: integer,
        the minimum number of operations to use parallel computation
    nReps: integer,
        number of replica of the construction
    ProbPoint: real
        between 0 and 1, probability of inclusing of a single point for each computation
    Subsets: list
        of column names (or column number). When specified a principal circle will be computed for each of the subsets specified.
    NumEdges: integer,
        the maximum nulber of edges
    Mode: integer,
        the energy computation mode
    FastSolve: boolean,
        should FastSolve be used when fitting the points to the data?
    ClusType: string,
        the type of cluster to use. It can gbe either "Sock" or "Fork".
    Currently fork clustering only works in Linux
    ICOver: string,
        initial condition overlap mode. This can be used to alter the default behaviour for the initial configuration of the
    principal circle
    DensityRadius: numeric,
        the radius used to estimate local density. This need to be set when ICOver is equal to "Density"
    AvoidSolitary: boolean,
        should configurations with "solitary nodes", i.e., nodes without associted points be discarded?
    FinalEnergy: string
        indicating the final elastic emergy associated with the configuration. Currently it can be "Base" or "Penalized"
    alpha: positive
        numeric, the value of the alpha parameter of the penalized elastic energy
    beta: positive
        numeric, the value of the beta parameter of the penalized elastic energy
    EmbPointProb: numeric between 0 and 1.
        If less than 1 point will be sampled at each iteration.
        EmbPointProb indicates the probability of using each points. This is an *experimental* feature, which may
        helps speeding up the computation if a large number of points is present.
    AdjustElasticMatrix: function
        a penalization function to adjust the elastic matrices after a configuration has been chosen (e.g., AdjustByConstant).
        If None (the default), no penalization will be used.
    AdjustVect: boolean
        vector keeping track of the nodes for which the elasticity parameters have been adjusted. When true for a node its elasticity parameters will not be adjusted.
    AdjustElasticMatrix_Initial: function
        a penalization function to adjust the elastic matrices of the initial configuration (e.g., AdjustByConstant).
        If None (the default), no penalization will be used.
    Lambda_Initial: real
        the lambda parameter used the construct the elastic matrix associted with ther initial configuration if needed.
        If None, the value of Lambda will be used.
    Mu_Initial: real,
        the mu parameter used the construct the elastic matrix associted with ther initial configuration if needed.
        If None, the value of Mu will be used.
    """

    # Define the initial configuration

    if ICOver is None:
        Configuration = "Circle"

    else:
        print(
            "WARNING : ICOver is currently ignored when constructing a circle"
        )
        Configuration = "Circle"

    if InitNodes < 3:
        print(
            "WARNING: The initial number of nodes must be at least 3. This"
            " will be fixed"
        )
        InitNodes = 3

    return elpigraph.computeElasticPrincipalGraphWithGrammars(
        X=X,
        NumNodes=NumNodes,
        NumEdges=NumEdges,
        InitNodes=InitNodes,
        Lambda=Lambda,
        Mu=Mu,
        GrammarOptimization=GrammarOptimization,
        GrowGrammars=np.array([["bisectedge"]]),
        ShrinkGrammars=np.array([]),
        MaxNumberOfIterations=MaxNumberOfIterations,
        TrimmingRadius=TrimmingRadius,
        eps=eps,
        Do_PCA=Do_PCA,
        InitNodePositions=InitNodePositions,
        InitEdges=InitEdges,
        AdjustVect=AdjustVect,
        ElasticMatrix=ElasticMatrix,
        Configuration=Configuration,
        CenterData=CenterData,
        ComputeMSEP=ComputeMSEP,
        verbose=verbose,
        ShowTimer=ShowTimer,
        ReduceDimension=ReduceDimension,
        # drawAccuracyComplexity = drawAccuracyComplexity,
        # drawPCAView = drawPCAView,
        # drawEnergy = drawEnergy,
        n_cores=n_cores,
        MinParOp=MinParOp,
        #                                              ClusType = ClusType,
        nReps=nReps,
        Subsets=Subsets,
        ProbPoint=ProbPoint,
        Mode=Mode,
        FinalEnergy=FinalEnergy,
        alpha=alpha,
        beta=beta,
        #                                              FastSolve = FastSolve,
        DensityRadius=DensityRadius,
        AvoidSolitary=AvoidSolitary,
        EmbPointProb=EmbPointProb,
        PointWeights=PointWeights,
        SampleIC=SampleIC,
        AvoidResampling=AvoidResampling,
        #                                              ParallelRep = ParallelRep,
        AdjustElasticMatrix=AdjustElasticMatrix,
        AdjustElasticMatrix_Initial=AdjustElasticMatrix_Initial,
        Lambda_Initial=Lambda_Initial,
        Mu_Initial=Mu_Initial,
        DisplayWarnings=DisplayWarnings,
        MaxSteps=MaxSteps,
        StoreGraphEvolution=StoreGraphEvolution,
        GPU=GPU,
        FixNodesAtPoints=FixNodesAtPoints,
        pseudotime=pseudotime,
        pseudotimeLambda=pseudotimeLambda,
        label=label,
        labelLambda=labelLambda,
        MaxNumberOfGraphCandidatesDict=MaxNumberOfGraphCandidatesDict,
    )


def computeElasticPrincipalTree(
    X,
    NumNodes,
    NumEdges=float("inf"),
    InitNodes=3,
    Lambda=0.01,
    Mu=0.1,
    GrammarOptimization=False,
    MaxSteps=float("inf"),
    GrammarOrder=["Grow", "Shrink"],
    MaxNumberOfIterations=10,
    TrimmingRadius=float("inf"),
    eps=0.01,
    Do_PCA=True,
    InitNodePositions=None,
    AdjustVect=None,
    ElasticMatrix=None,
    InitEdges=None,
    CenterData=True,
    ComputeMSEP=True,
    verbose=False,
    ShowTimer=False,
    ReduceDimension=None,
    # drawAccuracyComplexity = True,
    # drawPCAView = True,
    # drawEnergy = True,
    n_cores=1,
    # ClusType = "Sock",
    MinParOp=20,
    nReps=1,
    # ParallelRep = False,
    Subsets=list(),
    ProbPoint=1,
    Mode=1,
    FinalEnergy="Penalized",
    alpha=0,
    beta=0,
    # gamma = 0,
    # FastSolve = False,
    Configuration="Line",
    ICOver=None,
    DensityRadius=None,
    AvoidSolitary=False,
    EmbPointProb=1,
    PointWeights=None,
    SampleIC=True,
    AvoidResampling=True,
    AdjustElasticMatrix=None,
    AdjustElasticMatrix_Initial=None,
    Lambda_Initial=None,
    Mu_Initial=None,
    DisplayWarnings=False,
    StoreGraphEvolution=False,
    GPU=False,
    FixNodesAtPoints=[],
    pseudotime=None,
    pseudotimeLambda=0.01,
    label=None,
    labelLambda=0.01,
    MaxNumberOfGraphCandidatesDict={
        "AddNode2Node": float("inf"),
        "BisectEdge": float("inf"),
        "RemoveNode": float("inf"),
        "ShrinkEdge": float("inf"),
    },
):
    """
    Construct a principal elastic tree

    This function is a wrapper to the computeElasticPrincipalGraph function that constructs the appropriate initial graph and grammars
    when constructing a tree

    X: numerical 2D matrix,
        the n-by-m matrix with the position of n m-dimensional points
    NumNodes: integer,
        the number of nodes of the principal graph
    Lambda: real,
        the lambda parameter used the compute the elastic energy
    Mu: real,
        the lambda parameter used the compute the elastic energy
    InitNodes: integer,
        number of points to include in the initial graph
    MaxNumberOfIterations: integer,
        maximum number of steps to embed the nodes in the data
    TrimmingRadius: real,
        maximal distance of point from a node to affect its embedment
    eps: real,
        minimal relative change in the position of the nodes to stop embedment
    Do_PCA: boolean,
        should data and initial node positions be PCA trnasformed?
    InitNodePositions: numerical 2D matrix,
        the k-by-m matrix with k m-dimensional positions of the nodes
    in the initial step
    InitEdges: numerical 2D matrix,
        the e-by-2 matrix with e end-points of the edges connecting the nodes
    ElasticMatrix: numerical 2D matrix,
        the k-by-k elastic matrix
    CenterData: boolean,
        should data and initial node positions be centered?
    ComputeMSEP: boolean,
        should MSEP be computed when building the report?
    verbose: boolean,
        should debugging information be reported?
    ShowTimer: boolean,
        should the time to construct the graph be computed and reported for each step?
    ReduceDimension: integer vector,
        vector of principal components to retain when performing
        dimensionality reduction. If None all the components will be used
    drawAccuracyComplexity: boolean,
        should the accuracy VS complexity plot be reported?
    drawPCAView: boolean,
        should a 2D plot of the points and pricipal curve be dranw for the final configuration?
    drawEnergy: boolean,
        should changes of evergy VS the number of nodes be reported?
    n:.cores
        either an integer (indicating the number of cores to used for the creation of a cluster) or
    cluster structure returned, e.g., by makeCluster. If a cluster structure is used, all the nodes must contains X
    (this is done using clusterExport)
    MinParOP: integer,
        the minimum number of operations to use parallel computation
    nReps: integer,
        number of replica of the construction
    ProbPoint: real
        between 0 and 1, probability of inclusing of a single point for each computation
    Subsets: list
        of column names (or column number). When specified a principal tree will be computed for each of the subsets specified.
    NumEdges: integer,
        the maximum nulber of edges
    Mode: integer,
        the energy computation mode
    FastSolve: boolean,
        should FastSolve be used when fitting the points to the data?
    ClusType: string,
        the type of cluster to use. It can gbe either "Sock" or "Fork".
    Currently fork clustering only works in Linux
    ICOver: string,
        initial condition overlap mode. This can be used to alter the default behaviour for the initial configuration of the
    principal tree.
    DensityRadius: numeric,
        the radius used to estimate local density. This need to be set when ICOver is equal to "Density"
    AvoidSolitary: boolean,
        should configurations with "solitary nodes", i.e., nodes without associted points be discarded?
    FinalEnergy: string
        indicating the final elastic emergy associated with the configuration. Currently it can be "Base" or "Penalized"
    alpha: positive
        numeric, the value of the alpha parameter of the penalized elastic energy
    beta: positive
        numeric, the value of the beta parameter of the penalized elastic energy
    EmbPointProb: numeric between 0 and 1.
        If less than 1 point will be sampled at each iteration.
        EmbPointProb indicates the probability of using each points. This is an *experimental* feature, which may
        helps speeding up the computation if a large number of points is present.
    AdjustElasticMatrix: function
        a penalization function to adjust the elastic matrices after a configuration has been chosen (e.g., AdjustByConstant).
        If None (the default), no penalization will be used.
    AdjustVect: boolean
        vector keeping track of the nodes for which the elasticity parameters have been adjusted. When true for a node its elasticity parameters will not be adjusted.
    AdjustElasticMatrix_Initial: function
        a penalization function to adjust the elastic matrices of the initial configuration (e.g., AdjustByConstant).
        If None (the default), no penalization will be used.
    Lambda_Initial: real
        the lambda parameter used the construct the elastic matrix associted with ther initial configuration if needed.
        If None, the value of Lambda will be used.
    Mu_Initial: real,
        the mu parameter used the construct the elastic matrix associted with ther initial configuration if needed.
        If None, the value of Mu will be used.

    Return
    ------
    A list of principal graph structures containing the trees constructed during the different replica of the algorithm.
    If the number of replicas is larger than 1. The the final element of the list is the "average tree", which is constructed by
    fitting the coordinates of the nodes of the reconstructed trees
    """

    # Define the initial configuration

    if ICOver is None:
        Configuration = "Line"

    else:
        Configuration = ICOver

    return elpigraph.computeElasticPrincipalGraphWithGrammars(
        X=X,
        NumNodes=NumNodes,
        NumEdges=NumEdges,
        InitNodes=InitNodes,
        Lambda=Lambda,
        Mu=Mu,
        GrammarOptimization=GrammarOptimization,
        GrowGrammars=np.array(
            [["bisectedge", "addnode2node"], ["bisectedge", "addnode2node"]]
        ),
        ShrinkGrammars=np.array([["shrinkedge", "removenode"]]),
        MaxNumberOfIterations=MaxNumberOfIterations,
        TrimmingRadius=TrimmingRadius,
        eps=eps,
        Do_PCA=Do_PCA,
        InitNodePositions=InitNodePositions,
        InitEdges=InitEdges,
        AdjustVect=AdjustVect,
        ElasticMatrix=ElasticMatrix,
        Configuration=Configuration,
        CenterData=CenterData,
        ComputeMSEP=ComputeMSEP,
        verbose=verbose,
        ShowTimer=ShowTimer,
        ReduceDimension=ReduceDimension,
        # drawAccuracyComplexity = drawAccuracyComplexity,
        # drawPCAView = drawPCAView,
        # drawEnergy = drawEnergy,
        n_cores=n_cores,
        MinParOp=MinParOp,
        #                                              ClusType = ClusType,
        nReps=nReps,
        Subsets=Subsets,
        ProbPoint=ProbPoint,
        Mode=Mode,
        FinalEnergy=FinalEnergy,
        alpha=alpha,
        beta=beta,
        #                                              FastSolve = FastSolve,
        DensityRadius=DensityRadius,
        AvoidSolitary=AvoidSolitary,
        EmbPointProb=EmbPointProb,
        PointWeights=PointWeights,
        SampleIC=SampleIC,
        AvoidResampling=AvoidResampling,
        #                                              ParallelRep = ParallelRep,
        AdjustElasticMatrix=AdjustElasticMatrix,
        AdjustElasticMatrix_Initial=AdjustElasticMatrix_Initial,
        Lambda_Initial=Lambda_Initial,
        Mu_Initial=Mu_Initial,
        DisplayWarnings=DisplayWarnings,
        MaxSteps=MaxSteps,
        StoreGraphEvolution=StoreGraphEvolution,
        GPU=GPU,
        FixNodesAtPoints=FixNodesAtPoints,
        pseudotime=pseudotime,
        pseudotimeLambda=pseudotimeLambda,
        label=label,
        labelLambda=labelLambda,
        MaxNumberOfGraphCandidatesDict=MaxNumberOfGraphCandidatesDict,
    )


def computeElasticPrincipalCurve(
    X,
    NumNodes,
    NumEdges=float("inf"),
    InitNodes=3,
    Lambda=0.01,
    Mu=0.1,
    GrammarOptimization=False,
    MaxSteps=float("inf"),
    GrammarOrder=["Grow", "Shrink"],
    MaxNumberOfIterations=10,
    TrimmingRadius=float("inf"),
    eps=0.01,
    Do_PCA=True,
    InitNodePositions=None,
    AdjustVect=None,
    ElasticMatrix=None,
    InitEdges=None,
    CenterData=True,
    ComputeMSEP=True,
    verbose=False,
    ShowTimer=False,
    ReduceDimension=None,
    # drawAccuracyComplexity = True,
    # drawPCAView = True,
    # drawEnergy = True,
    n_cores=1,
    # ClusType = "Sock",
    MinParOp=20,
    nReps=1,
    # ParallelRep = False,
    Subsets=list(),
    ProbPoint=1,
    Mode=1,
    FinalEnergy="Base",
    alpha=0,
    beta=0,
    # gamma = 0,
    # FastSolve = False,
    Configuration="Line",
    ICOver=None,
    DensityRadius=None,
    AvoidSolitary=False,
    EmbPointProb=1,
    PointWeights=None,
    SampleIC=True,
    AvoidResampling=True,
    AdjustElasticMatrix=None,
    AdjustElasticMatrix_Initial=None,
    Lambda_Initial=None,
    Mu_Initial=None,
    DisplayWarnings=False,
    StoreGraphEvolution=False,
    GPU=False,
    FixNodesAtPoints=[],
    pseudotime=None,
    pseudotimeLambda=0.01,
    label=None,
    labelLambda=0.01,
    MaxNumberOfGraphCandidatesDict={
        "AddNode2Node": float("inf"),
        "BisectEdge": float("inf"),
        "RemoveNode": float("inf"),
        "ShrinkEdge": float("inf"),
    },
):

    """
    Construct a principal elastic curve

    This function is a wrapper to the computeElasticPrincipalGraph function that constructs the appropriate initial graph and grammars
    when constructing a curve

    X: numerical 2D matrix,
        the n-by-m matrix with the position of n m-dimensional points
    NumNodes: integer,
        the number of nodes of the principal graph
    Lambda: real,
        the lambda parameter used the compute the elastic energy
    Mu: real,
        the lambda parameter used the compute the elastic energy
    InitNodes: integer,
        number of points to include in the initial graph
    MaxNumberOfIterations: integer,
        maximum number of steps to embed the nodes in the data
    TrimmingRadius: real,
        maximal distance of point from a node to affect its embedment
    eps: real,
        minimal relative change in the position of the nodes to stop embedment
    Do_PCA: boolean,
        should data and initial node positions be PCA trnasformed?
    InitNodePositions: numerical 2D matrix,
        the k-by-m matrix with k m-dimensional positions of the nodes
    in the initial step
    InitEdges: numerical 2D matrix,
        the e-by-2 matrix with e end-points of the edges connecting the nodes
    ElasticMatrix: numerical 2D matrix,
        the k-by-k elastic matrix
    CenterData: boolean,
        should data and initial node positions be centered?
    ComputeMSEP: boolean,
        should MSEP be computed when building the report?
    verbose: boolean,
        should debugging information be reported?
    ShowTimer: boolean,
        should the time to construct the graph be computed and reported for each step?
    ReduceDimension: integer vector,
        vector of principal components to retain when performing
        dimensionality reduction. If None all the components will be used
    drawAccuracyComplexity: boolean,
        should the accuracy VS complexity plot be reported?
    drawPCAView: boolean,
        should a 2D plot of the points and pricipal curve be dranw for the final configuration?
    drawEnergy: boolean,
        should changes of evergy VS the number of nodes be reported?
    n:.cores
        either an integer (indicating the number of cores to used for the creation of a cluster) or
    cluster structure returned, e.g., by makeCluster. If a cluster structure is used, all the nodes must contains X
    (this is done using clusterExport)
    MinParOP: integer,
        the minimum number of operations to use parallel computation
    nReps: integer,
        number of replica of the construction
    ProbPoint: real
        between 0 and 1, probability of inclusing of a single point for each computation
    Subsets: list
        of column names (or column number). When specified a principal curve will be computed for each of the subsets specified.
    NumEdges: integer,
        the maximum nulber of edges
    Mode: integer,
        the energy computation mode
    FastSolve: boolean,
        should FastSolve be used when fitting the points to the data?
    ClusType: string,
        the type of cluster to use. It can gbe either "Sock" or "Fork".
    Currently fork clustering only works in Linux
    ICOver: string,
        initial condition overlap mode. This can be used to alter the default behaviour for the initial configuration of the
    principal curve.
    DensityRadius: numeric,
        the radius used to estimate local density. This need to be set when ICOver is equal to "Density"
    AvoidSolitary: boolean,
        should configurations with "solitary nodes", i.e., nodes without associted points be discarded?
    FinalEnergy: string
        indicating the final elastic emergy associated with the configuration. Currently it can be "Base" or "Penalized"
    alpha: positive
        numeric, the value of the alpha parameter of the penalized elastic energy
    beta: positive
        numeric, the value of the beta parameter of the penalized elastic energy
    EmbPointProb: numeric between 0 and 1.
        If less than 1 point will be sampled at each iteration.
        EmbPointProb indicates the probability of using each points. This is an *experimental* feature, which may
        helps speeding up the computation if a large number of points is present.
    AdjustElasticMatrix: function
        a penalization function to adjust the elastic matrices after a configuration has been chosen (e.g., AdjustByConstant).
        If None (the default), no penalization will be used.
    AdjustVect: boolean
        vector keeping track of the nodes for which the elasticity parameters have been adjusted. When true for a node its elasticity parameters will not be adjusted.
    AdjustElasticMatrix_Initial: function
        a penalization function to adjust the elastic matrices of the initial configuration (e.g., AdjustByConstant).
        If None (the default), no penalization will be used.
    Lambda_Initial: real
        the lambda parameter used the construct the elastic matrix associted with ther initial configuration if needed.
        If None, the value of Lambda will be used.
    Mu_Initial: real,
        the mu parameter used the construct the elastic matrix associted with ther initial configuration if needed.
        If None, the value of Mu will be used. boolean,
        should parallel execution be performed on the sampling instead of the the grammar evaluations?
    AvoidResampling: booleand,
        should the sampling of initial conditions avoid reselecting the same points
    (or points neighbors if DensityRadius is specified)?
    SampleIC: boolean,
        should the initial configuration be considered on the sampled points when applicable?

    Return
    ------
    A list of principal graph strucutures containing the curves constructed during the different replica of the algorithm.
    If the number of replicas is larger than 1. The the final element of the list is the "average curve", which is constructed by
    fitting the coordinates of the nodes of the reconstructed curves
    """
    # Define the initial configuration
    if ICOver is None:
        Configuration = "Line"

    else:
        Configuration = ICOver

    return elpigraph.computeElasticPrincipalGraphWithGrammars(
        X=X,
        NumNodes=NumNodes,
        NumEdges=NumEdges,
        InitNodes=InitNodes,
        Lambda=Lambda,
        Mu=Mu,
        GrammarOptimization=GrammarOptimization,
        GrowGrammars=np.array([["bisectedge"]]),
        ShrinkGrammars=np.array([]),
        MaxNumberOfIterations=MaxNumberOfIterations,
        TrimmingRadius=TrimmingRadius,
        eps=eps,
        Do_PCA=Do_PCA,
        InitNodePositions=InitNodePositions,
        InitEdges=InitEdges,
        AdjustVect=AdjustVect,
        ElasticMatrix=ElasticMatrix,
        Configuration=Configuration,
        CenterData=CenterData,
        ComputeMSEP=ComputeMSEP,
        verbose=verbose,
        ShowTimer=ShowTimer,
        ReduceDimension=ReduceDimension,
        # drawAccuracyComplexity = drawAccuracyComplexity,
        # drawPCAView = drawPCAView,
        # drawEnergy = drawEnergy,
        n_cores=n_cores,
        MinParOp=MinParOp,
        #                                              ClusType = ClusType,
        nReps=nReps,
        Subsets=Subsets,
        ProbPoint=ProbPoint,
        Mode=Mode,
        FinalEnergy=FinalEnergy,
        alpha=alpha,
        beta=beta,
        #                                              FastSolve = FastSolve,
        DensityRadius=DensityRadius,
        AvoidSolitary=AvoidSolitary,
        EmbPointProb=EmbPointProb,
        PointWeights=PointWeights,
        SampleIC=SampleIC,
        AvoidResampling=AvoidResampling,
        #                                              ParallelRep = ParallelRep,
        AdjustElasticMatrix=AdjustElasticMatrix,
        AdjustElasticMatrix_Initial=AdjustElasticMatrix_Initial,
        Lambda_Initial=Lambda_Initial,
        Mu_Initial=Mu_Initial,
        DisplayWarnings=DisplayWarnings,
        MaxSteps=MaxSteps,
        StoreGraphEvolution=StoreGraphEvolution,
        GPU=GPU,
        FixNodesAtPoints=FixNodesAtPoints,
        pseudotime=pseudotime,
        pseudotimeLambda=pseudotimeLambda,
        label=label,
        labelLambda=labelLambda,
        MaxNumberOfGraphCandidatesDict=MaxNumberOfGraphCandidatesDict,
    )


def fineTuneBR(
    X,
    NumNodes,
    NumEdges=float("inf"),
    InitNodes=3,
    Lambda=0.01,
    Mu=0.1,
    GrammarOptimization=False,
    MaxSteps=100,
    GrammarOrder=["Grow", "Shrink"],
    MaxNumberOfIterations=10,
    TrimmingRadius=float("inf"),
    eps=0.01,
    Do_PCA=True,
    InitNodePositions=None,
    AdjustVect=None,
    ElasticMatrix=None,
    InitEdges=None,
    CenterData=True,
    ComputeMSEP=True,
    verbose=False,
    ShowTimer=False,
    ReduceDimension=None,
    # drawAccuracyComplexity = True,
    # drawPCAView = True,
    # drawEnergy = True,
    n_cores=1,
    # ClusType = "Sock",
    MinParOp=20,
    nReps=1,
    # ParallelRep = False,
    Subsets=list(),
    ProbPoint=1,
    Mode=1,
    FinalEnergy="Base",
    alpha=0,
    beta=0,
    # gamma = 0,
    # FastSolve = False,
    Configuration="Line",
    ICOver=None,
    DensityRadius=None,
    AvoidSolitary=False,
    EmbPointProb=1,
    PointWeights=None,
    SampleIC=True,
    AvoidResampling=True,
    AdjustElasticMatrix=None,
    AdjustElasticMatrix_Initial=None,
    Lambda_Initial=None,
    Mu_Initial=None,
    DisplayWarnings=False,
    StoreGraphEvolution=False,
    GPU=False,
    FixNodesAtPoints=[],
    pseudotime=None,
    pseudotimeLambda=0.01,
    label=None,
    labelLambda=0.01,
    MaxNumberOfGraphCandidatesDict={
        "AddNode2Node": float("inf"),
        "BisectEdge": float("inf"),
        "RemoveNode": float("inf"),
        "ShrinkEdge": float("inf"),
    },
):

    """
    Expand the nodes around a branching point.
    This function is a wrapper to the computeElasticPrincipalGraph function that construct the appropriate initial graph and grammars
    when increasing the nume number around the branching point

    X: numerical 2D matrix,
        the n-by-m matrix with the position of n m-dimensional points
    NumNodes: integer,
        the number of nodes of the principal graph
    Lambda: real,
        the lambda parameter used the compute the elastic energy
    Mu: real,
        the lambda parameter used the compute the elastic energy
    InitNodes: integer,
        number of points to include in the initial graph
    MaxNumberOfIterations: integer,
        maximum number of steps to embed the nodes in the data
    TrimmingRadius: real,
        maximal distance of point from a node to affect its embedment
    eps: real,
        minimal relative change in the position of the nodes to stop embedment
    Do_PCA: boolean,
        should data and initial node positions be PCA trnasformed?
    InitNodePositions: numerical 2D matrix,
        the k-by-m matrix with k m-dimensional positions of the nodes
    in the initial step
    InitEdges: numerical 2D matrix,
        the e-by-2 matrix with e end-points of the edges connecting the nodes
    ElasticMatrix: numerical 2D matrix,
        the k-by-k elastic matrix
    CenterData: boolean,
        should data and initial node positions be centered?
    ComputeMSEP: boolean,
        should MSEP be computed when building the report?
    verbose: boolean,
        should debugging information be reported?
    ShowTimer: boolean,
        should the time to construct the graph be computed and reported for each step?
    ReduceDimension: integer vector,
        vector of principal components to retain when performing
        dimensionality reduction. If None all the components will be used
    drawAccuracyComplexity: boolean,
        should the accuracy VS complexity plot be reported?
    drawPCAView: boolean,
        should a 2D plot of the points and pricipal curve be dranw for the final configuration?
    drawEnergy: boolean,
        should changes of evergy VS the number of nodes be reported?
    n:.cores
        either an integer (indicating the number of cores to used for the creation of a cluster) or
    cluster structure returned, e.g., by makeCluster. If a cluster structure is used, all the nodes must contains X
    (this is done using clusterExport)
    MinParOP: integer,
        the minimum number of operations to use parallel computation
    nReps: integer,
        number of replica of the construction
    ProbPoint: real
        between 0 and 1, probability of inclusing of a single point for each computation
    Subsets: list
        of column names (or column number). When specified a principal curve will be computed for each of the subsets specified.
    NumEdges: integer,
        the maximum nulber of edges
    Mode: integer,
        the energy computation mode
    FastSolve: boolean,
        should FastSolve be used when fitting the points to the data?
    ClusType: string,
        the type of cluster to use. It can gbe either "Sock" or "Fork".
    Currently fork clustering only works in Linux
    ICOver: string,
        initial condition overlap mode. This can be used to alter the default behaviour for the initial configuration of the
    principal curve.
    DensityRadius: numeric,
        the radius used to estimate local density. This need to be set when ICOver is equal to "Density"
    AvoidSolitary: boolean,
        should configurations with "solitary nodes", i.e., nodes without associted points be discarded?
    FinalEnergy: string
        indicating the final elastic emergy associated with the configuration. Currently it can be "Base" or "Penalized"
    alpha: positive
        numeric, the value of the alpha parameter of the penalized elastic energy
    beta: positive
        numeric, the value of the beta parameter of the penalized elastic energy
    EmbPointProb: numeric between 0 and 1.
        If less than 1 point will be sampled at each iteration.
        EmbPointProb indicates the probability of using each points. This is an *experimental* feature, which may
        helps speeding up the computation if a large number of points is present.
    AdjustElasticMatrix: function
        a penalization function to adjust the elastic matrices after a configuration has been chosen (e.g., AdjustByConstant).
        If None (the default), no penalization will be used.
    AdjustVect: boolean
        vector keeping track of the nodes for which the elasticity parameters have been adjusted. When true for a node its elasticity parameters will not be adjusted.
    AdjustElasticMatrix_Initial: function
        a penalization function to adjust the elastic matrices of the initial configuration (e.g., AdjustByConstant).
        If None (the default), no penalization will be used.
    Lambda_Initial: real
        the lambda parameter used the construct the elastic matrix associted with ther initial configuration if needed.
        If None, the value of Lambda will be used.
    Mu_Initial: real,
        the mu parameter used the construct the elastic matrix associted with ther initial configuration if needed.
        If None, the value of Mu will be used.
    SampleIC: boolean,
        should the initial configuration be considered on the sampled points when applicable?

    Return
    ------
    A list of principal graph structures containing the curves constructed during the different replica of the algorithm.
    If the number of replicas is larger than 1. The the final element of the list is the "average curve", which is constructed by
    fitting the coordinates of the nodes of the reconstructed curves
    """
    # Define the initial configuration

    if ICOver is None:
        Configuration = "Line"

    else:
        Configuration = ICOver

    if NumNodes > len(InitNodePositions):
        GrammarOptimization = False
        GrowGrammars = np.array([["bisectedge_3"]])
        ShrinkGrammars = np.array([["shrinkedge_3"]])
        GrammarOrder = ["Grow", "Shrink", "Grow"]
    else:
        GrammarOptimization = True
        GrowGrammars = np.array([["bisectedge_3"]])
        ShrinkGrammars = np.array([["shrinkedge_3"]])
        GrammarOrder = ["Shrink", "Grow"]

    return elpigraph.computeElasticPrincipalGraphWithGrammars(
        X=X,
        NumNodes=NumNodes,
        NumEdges=NumEdges,
        InitNodes=InitNodes,
        Lambda=Lambda,
        Mu=Mu,
        GrowGrammars=GrowGrammars,
        ShrinkGrammars=ShrinkGrammars,
        GrammarOrder=GrammarOrder,
        GrammarOptimization=GrammarOptimization,
        MaxSteps=MaxSteps,
        MaxNumberOfIterations=MaxNumberOfIterations,
        TrimmingRadius=TrimmingRadius,
        eps=eps,
        Do_PCA=Do_PCA,
        InitNodePositions=InitNodePositions,
        InitEdges=InitEdges,
        AdjustVect=AdjustVect,
        ElasticMatrix=ElasticMatrix,
        Configuration=Configuration,
        CenterData=CenterData,
        ComputeMSEP=ComputeMSEP,
        verbose=verbose,
        ShowTimer=ShowTimer,
        ReduceDimension=ReduceDimension,
        # drawAccuracyComplexity = drawAccuracyComplexity,
        # drawPCAView = drawPCAView,
        # drawEnergy = drawEnergy,
        n_cores=n_cores,
        MinParOp=MinParOp,
        #                                              ClusType = ClusType,
        nReps=nReps,
        Subsets=Subsets,
        ProbPoint=ProbPoint,
        Mode=Mode,
        FinalEnergy=FinalEnergy,
        alpha=alpha,
        beta=beta,
        #                                              FastSolve = FastSolve,
        DensityRadius=DensityRadius,
        AvoidSolitary=AvoidSolitary,
        EmbPointProb=EmbPointProb,
        PointWeights=PointWeights,
        SampleIC=SampleIC,
        AvoidResampling=AvoidResampling,
        #                                              ParallelRep = ParallelRep,
        AdjustElasticMatrix=AdjustElasticMatrix,
        AdjustElasticMatrix_Initial=AdjustElasticMatrix_Initial,
        Lambda_Initial=Lambda_Initial,
        Mu_Initial=Mu_Initial,
        DisplayWarnings=DisplayWarnings,
        StoreGraphEvolution=StoreGraphEvolution,
        GPU=GPU,
        FixNodesAtPoints=FixNodesAtPoints,
        pseudotime=pseudotime,
        pseudotimeLambda=pseudotimeLambda,
        label=label,
        labelLambda=labelLambda,
        MaxNumberOfGraphCandidatesDict=MaxNumberOfGraphCandidatesDict,
    )


def GrowLeaves(
    X,
    NumNodes,
    NumEdges=float("inf"),
    InitNodes=3,
    Lambda=0.01,
    Mu=0.1,
    MaxSteps=100,
    MaxNumberOfIterations=10,
    TrimmingRadius=float("inf"),
    eps=0.01,
    Do_PCA=True,
    InitNodePositions=None,
    AdjustVect=None,
    ElasticMatrix=None,
    InitEdges=None,
    CenterData=True,
    ComputeMSEP=True,
    verbose=False,
    ShowTimer=False,
    ReduceDimension=None,
    # drawAccuracyComplexity = True,
    # drawPCAView = True,
    # drawEnergy = True,
    n_cores=1,
    # ClusType = "Sock",
    MinParOp=20,
    nReps=1,
    # ParallelRep = False,
    Subsets=list(),
    ProbPoint=1,
    Mode=1,
    FinalEnergy="Base",
    alpha=0,
    beta=0,
    # gamma = 0,
    # FastSolve = False,
    Configuration="Line",
    ICOver=None,
    DensityRadius=None,
    AvoidSolitary=False,
    EmbPointProb=1,
    PointWeights=None,
    SampleIC=True,
    AvoidResampling=True,
    AdjustElasticMatrix=None,
    AdjustElasticMatrix_Initial=None,
    Lambda_Initial=None,
    Mu_Initial=None,
    DisplayWarnings=False,
    StoreGraphEvolution=False,
    GPU=False,
    FixNodesAtPoints=[],
    pseudotime=None,
    pseudotimeLambda=0.01,
    label=None,
    labelLambda=0.01,
    MaxNumberOfGraphCandidatesDict={
        "AddNode2Node": float("inf"),
        "BisectEdge": float("inf"),
        "RemoveNode": float("inf"),
        "ShrinkEdge": float("inf"),
    },
):

    """
    Extend the leaves of a graph

    This function is a wrapper to the computeElasticPrincipalGraph function that construct the appropriate initial graph and grammars
    when increasing the nume number around the branching point

    X: numerical 2D matrix,
        the n-by-m matrix with the position of n m-dimensional points
    NumNodes: integer,
        the number of nodes of the principal graph
    Lambda: real,
        the lambda parameter used the compute the elastic energy
    Mu: real,
        the lambda parameter used the compute the elastic energy
    InitNodes: integer,
        number of points to include in the initial graph
    MaxNumberOfIterations: integer,
        maximum number of steps to embed the nodes in the data
    TrimmingRadius: real,
        maximal distance of point from a node to affect its embedment
    eps: real,
        minimal relative change in the position of the nodes to stop embedment
    Do_PCA: boolean,
        should data and initial node positions be PCA trnasformed?
    InitNodePositions: numerical 2D matrix,
        the k-by-m matrix with k m-dimensional positions of the nodes
    in the initial step
    InitEdges: numerical 2D matrix,
        the e-by-2 matrix with e end-points of the edges connecting the nodes
    ElasticMatrix: numerical 2D matrix,
        the k-by-k elastic matrix
    CenterData: boolean,
        should data and initial node positions be centered?
    ComputeMSEP: boolean,
        should MSEP be computed when building the report?
    verbose: boolean,
        should debugging information be reported?
    ShowTimer: boolean,
        should the time to construct the graph be computed and reported for each step?
    ReduceDimension: integer vector,
        vector of principal components to retain when performing
        dimensionality reduction. If None all the components will be used
    drawAccuracyComplexity: boolean,
        should the accuracy VS complexity plot be reported?
    drawPCAView: boolean,
        should a 2D plot of the points and pricipal curve be dranw for the final configuration?
    drawEnergy: boolean,
        should changes of evergy VS the number of nodes be reported?
    n:.cores
        either an integer (indicating the number of cores to used for the creation of a cluster) or
    cluster structure returned, e.g., by makeCluster. If a cluster structure is used, all the nodes must contains X
    (this is done using clusterExport)`
    MinParOP: integer,
        the minimum number of operations to use parallel computation
    nReps: integer,
        number of replica of the construction
    ProbPoint: real
        between 0 and 1, probability of inclusing of a single point for each computation
    Subsets: list
        of column names (or column number). When specified a principal curve will be computed for each of the subsets specified.
    NumEdges: integer,
        the maximum nulber of edges
    Mode: integer,
        the energy computation mode
    FastSolve: boolean,
        should FastSolve be used when fitting the points to the data?
    ClusType: string,
        the type of cluster to use. It can gbe either "Sock" or "Fork".
    Currently fork clustering only works in Linux
    ICOver: string,
        initial condition overlap mode. This can be used to alter the default behaviour for the initial configuration of the
    principal curve.
    DensityRadius: numeric,
        the radius used to estimate local density. This need to be set when ICOver is equal to "Density"
    AvoidSolitary: boolean,
        should configurations with "solitary nodes", i.e., nodes without associted points be discarded?
    FinalEnergy: string
        indicating the final elastic emergy associated with the configuration. Currently it can be "Base" or "Penalized"
    alpha: positive
        numeric, the value of the alpha parameter of the penalized elastic energy
    beta: positive
        numeric, the value of the beta parameter of the penalized elastic energy
    EmbPointProb: numeric between 0 and 1.
        If less than 1 point will be sampled at each iteration.
        EmbPointProb indicates the probability of using each points. This is an *experimental* feature, which may
        helps speeding up the computation if a large number of points is present.
    AdjustElasticMatrix: function
        a penalization function to adjust the elastic matrices after a configuration has been chosen (e.g., AdjustByConstant).
        If None (the default), no penalization will be used.
    AdjustVect: boolean
        vector keeping track of the nodes for which the elasticity parameters have been adjusted. When true for a node its elasticity parameters will not be adjusted.
    AdjustElasticMatrix_Initial: function
        a penalization function to adjust the elastic matrices of the initial configuration (e.g., AdjustByConstant).
        If None (the default), no penalization will be used.
    Lambda_Initial: real
        the lambda parameter used the construct the elastic matrix associted with ther initial configuration if needed.
        If None, the value of Lambda will be used.
    Mu_Initial: real,
        the mu parameter used the construct the elastic matrix associted with ther initial configuration if needed.
        If None, the value of Mu will be used.
    SampleIC: boolean,
        should the initial configuration be considered on the sampled points when applicable?

    Return
    ------
    A list of principal graph structures containing the curves constructed during the different replica of the algorithm.
    If the number of replicas is larger than 1. The final element of the list is the "average curve", which is constructed by
    fitting the coordinates of the nodes of the reconstructed curves
    """

    # Define the initial configuration

    if ICOver is None:
        Configuration = "Line"

    else:
        Configuration = ICOver

    return elpigraph.computeElasticPrincipalGraphWithGrammars(
        X=X,
        NumNodes=NumNodes,
        NumEdges=NumEdges,
        InitNodes=InitNodes,
        Lambda=Lambda,
        Mu=Mu,
        GrowGrammars=np.array([["addnode2node_1"]]),
        ShrinkGrammars=np.array([]),
        GrammarOrder=["Grow"],
        GrammarOptimization=False,
        MaxSteps=MaxSteps,
        MaxNumberOfIterations=MaxNumberOfIterations,
        TrimmingRadius=TrimmingRadius,
        eps=eps,
        Do_PCA=Do_PCA,
        InitNodePositions=InitNodePositions,
        InitEdges=InitEdges,
        AdjustVect=AdjustVect,
        ElasticMatrix=ElasticMatrix,
        Configuration=Configuration,
        CenterData=CenterData,
        ComputeMSEP=ComputeMSEP,
        verbose=verbose,
        ShowTimer=ShowTimer,
        ReduceDimension=ReduceDimension,
        # drawAccuracyComplexity = drawAccuracyComplexity,
        # drawPCAView = drawPCAView,
        # drawEnergy = drawEnergy,
        n_cores=n_cores,
        MinParOp=MinParOp,
        #                                              ClusType = ClusType,
        nReps=nReps,
        Subsets=Subsets,
        ProbPoint=ProbPoint,
        Mode=Mode,
        FinalEnergy=FinalEnergy,
        alpha=alpha,
        beta=beta,
        #                                              FastSolve = FastSolve,
        DensityRadius=DensityRadius,
        AvoidSolitary=AvoidSolitary,
        EmbPointProb=EmbPointProb,
        PointWeights=PointWeights,
        SampleIC=SampleIC,
        AvoidResampling=AvoidResampling,
        #                                              ParallelRep = ParallelRep,
        AdjustElasticMatrix=AdjustElasticMatrix,
        AdjustElasticMatrix_Initial=AdjustElasticMatrix_Initial,
        Lambda_Initial=Lambda_Initial,
        Mu_Initial=Mu_Initial,
        DisplayWarnings=DisplayWarnings,
        StoreGraphEvolution=StoreGraphEvolution,
        GPU=GPU,
        FixNodesAtPoints=FixNodesAtPoints,
        pseudotime=pseudotime,
        pseudotimeLambda=pseudotimeLambda,
        label=label,
        labelLambda=labelLambda,
        MaxNumberOfGraphCandidatesDict=MaxNumberOfGraphCandidatesDict,
    )


def generateInitialConfiguration(
    X,
    Nodes,
    Configuration="Line",
    DensityRadius=None,
    MaxPoints=10000,
    PCADensity=True,
    CenterDataDensity=True,
    verbose=False,
    FixNodesAtPoints=[],
):
    """
    Produce an initial graph with a given structure

    X: numerical 2D matrix,
        the n-by-m matrix with the position of n m-dimensional points
    Configuration: string,
        type of graph to return. It should be one of the following values:
        Line:
            Points are placed on the 1st principal component between mean-sd and mean+sd
        Circle:
            Points are placed on the the plane induced by the 1st and 2nd principal components.
            In both dimensions they are placed between mean-sd and mean+sd
        Density:
            Two points are selected randomly in the neighborhood of one of the points with
            the largest number of neighbour points
        DensityProb:
            Two points are selected randomly in the neighborhood of one of a point randomly
            selected with a probability proportional to the number of neighbour points}
        Random:
            Two points are returned. The first is selected at random, and the second is selected with
            a probability inversely proportional to thr distance to the 1st point selected

    Nodes: integer,
        number of nodes of the graph
    DensityRadius: numeric,
        the radius used to estimate local density. This need to be set when Configuration is equal to "Density"
    MaxPoints: integer,
        the maximum number of points for which the local density will be estimated. If the number of data points is
        larger than MaxPoints, a subset of the original points will be sampled
    PCADensity: boolean,
        should PCA be applied to the data before computing the most dense area
    """
    DONE = False

    if Configuration == "Line":
        # Chain of nodes along the first principal component direction
        if verbose:
            print("Creating a chain in the 1st PC with", Nodes, "nodes")
        mv = X.mean(axis=0)
        data_centered = X - mv
        PC1, explainedVarianceRatio, U, S, Vt = TruncPCA(
            data_centered, n_components=1
        )
        # Vt = np.abs(Vt)
        mn = np.mean(PC1)
        st = np.std(PC1, ddof=1)
        NodeP = np.dot(np.linspace(mn - st, mn + st, Nodes)[:, None], Vt)
        NodePositions = NodeP + mv[None]

        # Creating edges
        edges = np.vstack((np.arange(Nodes - 1), np.arange(1, Nodes))).T
        DONE = True

    if Configuration == "Circle":
        # Chain of nodes along the first principal component direction
        if verbose:
            print(
                "Creating a circle in the plane induced by the 1st and 2nd PCs"
                " with",
                Nodes,
                "nodes",
            )

        if X.shape[1] < 3:
            if CenterDataDensity:
                mv = X.mean(axis=0)
                data_centered = X - mv
            else:
                data_centered = X
            vglobal, PCAdata, explainedVariances = PCA(data_centered)
            Vt = vglobal.T
        else:
            if CenterDataDensity:
                mv = X.mean(axis=0)
                data_centered = X - mv
            else:
                data_centered = X
            PCAdata, explainedVarianceRatio, U, S, Vt = TruncPCA(
                data_centered, n_components=2
            )

        Nodes_X = (
            np.cos(np.linspace(0, 2 * np.pi, Nodes + 1))
            * np.std(PCAdata[:, 0], ddof=1)
        )[:, None]
        Nodes_Y = (
            np.sin(np.linspace(0, 2 * np.pi, Nodes + 1))
            * np.std(PCAdata[:, 1], ddof=1)
        )[:, None]

        NodePositions = (
            np.dot(np.concatenate((Nodes_X[1:], Nodes_Y[1:]), axis=1), abs(Vt))
            + mv[None]
        )
        # Creating edges
        edges = np.vstack((np.arange(Nodes - 1), np.arange(1, Nodes))).T
        edges = np.concatenate((edges, np.array([[int(Nodes - 1), 0]])))

        DONE = True

    if Configuration == "Random":

        # Starting from Random Points in the data
        if verbose:
            print(
                "Creating a line between two random points of the data. The"
                " points will have at most a distance DensityRadius if the"
                " parameter is specified"
            )

        ID1 = np.random.choice(len(X))

        Dist = PartialDistance(
            X[
                [ID1],
            ],
            X,
        ).flatten()
        with np.errstate(divide="ignore"):
            Probs = 1 / (Dist)
        if DensityRadius:
            Probs[Dist > DensityRadius] = 0

        Probs[np.isinf(Probs)] = 0

        ID2 = np.random.choice(len(X), p=Probs / sum(Probs))

        NodePositions = X[np.array([ID1, ID2]), :]

        # Creating edges
        edges = np.array([[0, 1]])

        DONE = True

    if Configuration == "RandomSpace":

        # Starting from Random Points in the data
        if verbose:
            print(
                "Creating a line between a point randomly chosen uniformily in"
                " the space of points and one of its neighbours. The points"
                " will have at most a distance DensityRadius if the parameter"
                " is specified"
            )
        from sklearn.cluster import KMeans

        KM = KMeans(n_clusters=10).fit(X)
        RandClus = np.random.choice(10)
        ID1 = np.random.choice(np.where(KM.labels_ == RandClus)[0])

        Dist = PartialDistance(
            X[
                [ID1],
            ],
            X,
        ).flatten()

        with np.errstate(divide="ignore"):
            Probs = 1 / (Dist)
        if DensityRadius:
            Probs[Dist > DensityRadius] = 0
        Probs[np.isinf(Probs)] = 0

        ID2 = np.random.choice(len(X), p=Probs / sum(Probs))

        NodePositions = X[np.array([ID1, ID2]), :]

        # Creating edges
        edges = np.array([[0, 1]])

        DONE = True

    if Configuration == "Density":

        if DensityRadius is None:
            raise ValueError(
                "DensityRadius needs to be specified for density-dependent"
                " initialization!"
            )

        # Starting from Random Points in the data
        if verbose:
            print("Creating a line in the densest part of the data")

        if PCADensity:
            if CenterDataDensity:
                mv = X.mean(axis=0)
                data_centered = X - mv
            else:
                data_centered = X
            vglobal, tX_PCA, explainedVariances = PCA(data_centered)
            tX = tX_PCA
        else:
            tX = X

        if len(tX) > MaxPoints:
            if verbose:
                print(
                    "Too many points, a subset of",
                    MaxPoints,
                    "will be sampled",
                )

            SampledIdxs = np.random.choice(len(tX), size=MaxPoints)

            PartStruct = PartialDistance(
                tX[
                    SampledIdxs,
                ],
                tX[
                    SampledIdxs,
                ],
            )
            PointsInNei = np.sum(PartStruct < DensityRadius, axis=1)

            if max(PointsInNei) < 2:
                raise ValueError(
                    "DensityRadius too small (Not enough points found in the"
                    " neighborhood)"
                )

            IdMax = np.argmax(PointsInNei)

            NodePositions = X[
                SampledIdxs[
                    np.random.choice(
                        np.where(
                            PartStruct[
                                IdMax,
                            ]
                            < DensityRadius
                        )[0],
                        size=2,
                    )
                ],
                :,
            ]

            # Creating edges
            edges = np.array([[0, 1]])

            DONE = True

        else:
            PartStruct = PartialDistance(tX, tX)
            PointsInNei = np.sum(PartStruct < DensityRadius, axis=1)

            if max(PointsInNei) < 2:
                raise ValueError(
                    "DensityRadius too small (Not enough points found in the"
                    " neighborhood)"
                )

            IdMax = np.argmax(PointsInNei)
            NodePositions = X[
                np.random.choice(
                    np.where(
                        PartStruct[
                            IdMax,
                        ]
                        < DensityRadius
                    )[0],
                    size=2,
                ),
                :,
            ]

            # Creating edges
            edges = np.array([[0, 1]])

            DONE = True

    if Configuration == "DensityProb":

        if DensityRadius is None:
            raise ValueError(
                "DensityRadius need to be specified for density-dependent"
                " inizialization!"
            )

        # Starting from Random Points in the data
        if verbose:
            print(
                "Creating a line a part of the data, chosen probabilistically"
                " by its density. DensityRadius needs to be specified!"
            )

        if PCADensity:
            if CenterDataDensity:
                mv = X.mean(axis=0)
                data_centered = X - mv
            else:
                data_centered = X
            vglobal, tX_PCA, explainedVariances = PCA(data_centered)
            tX = tX_PCA
        else:
            tX = X

        if len(tX) > MaxPoints:
            if verbose:
                print(
                    "Too many points, a subset of",
                    MaxPoints,
                    "will be sampled",
                )

            SampledIdxs = np.random.choice(len(tX), size=MaxPoints)

            PartStruct = PartialDistance(
                tX[
                    SampledIdxs,
                ],
                tX[
                    SampledIdxs,
                ],
            )
            PointsInNei = np.sum(PartStruct < DensityRadius, axis=1)

            if max(PointsInNei) < 2:
                raise ValueError(
                    "DensityRadius too small (Not enough points found in the"
                    " neighborhood))!!"
                )

            PointsInNei[PointsInNei == 1] = 0
            IdMax = np.random.choice(
                len(PointsInNei), p=PointsInNei / sum(PointsInNei)
            )

            NodePositions = X[
                np.random.choice(
                    np.where(
                        PartStruct[
                            IdMax,
                        ]
                        < DensityRadius
                    )[0],
                    size=2,
                ),
                :,
            ]

            # Creating edges
            edges = np.array([[0, 1]])

            DONE = True

        else:
            PartStruct = PartialDistance(tX, tX)
            PointsInNei = np.sum(PartStruct < DensityRadius, axis=1)

            if max(PointsInNei) < 2:
                raise ValueError(
                    "DensityRadius too small (Not enough points found in the"
                    " neighborhood))!!"
                )

            PointsInNei[PointsInNei == 1] = 0
            IdMax = np.random.choice(
                len(PointsInNei), p=PointsInNei / sum(PointsInNei)
            )

            NodePositions = X[
                np.random.choice(
                    np.where(
                        PartStruct[
                            IdMax,
                        ]
                        < DensityRadius
                    )[0],
                    size=2,
                ),
                :,
            ]

            # Creating edges
            edges = np.array([[0, 1]])

            DONE = True

    if DONE:
        return dict(NodePositions=NodePositions, Edges=edges)
    else:
        raise ValueError("Unsupported configuration!")



import numpy as np


try:
    import cupy
except:
    pass
import multiprocessing as mp
import datetime
import time
import copy
from .PCA import PCA, TruncPCA, PCA_gpu, TruncSVD_gpu
from .core import (
    PrimitiveElasticGraphEmbedment,
    PrimitiveElasticGraphEmbedment_v2,
    PrimitiveElasticGraphEmbedment_cp,
    PrimitiveElasticGraphEmbedment_cp_v2,
    PartitionData,
    PartitionData_cp,
    MakeUniformElasticMatrix,
    Encode2ElasticMatrix,
    DecodeElasticMatrix,
)
from .grammar_operations import (
    ApplyOptimalGraphGrammarOperation,
    ApplyOptimalGraphGrammarOperation_v2,
    ApplyOptimalGraphGrammarOperation_v3,
)

from .reporting import ReportOnPrimitiveGraphEmbedment


def isnumeric(obj):
    try:
        obj + 0
        return True
    except:
        return False


def ElPrincGraph(
    X,
    Lambda,
    Mu,
    ElasticMatrix,
    NodePositions,
    AdjustVect,
    NumNodes=100,
    NumEdges=float("inf"),
    verbose=False,
    n_cores=1,
    #                 ClusType = "Sock",
    MinParOp=20,
    CompileReport=True,
    ShowTimer=False,
    ComputeMSEP=True,
    FinalEnergy="Base",
    alpha=0,
    beta=0,
    Mode=1,
    MaxBlockSize=100000000,
    MaxNumberOfIterations=10,
    MaxFailedOperations=float("inf"),
    MaxSteps=float("inf"),
    GrammarOptimization=True,
    eps=0.01,
    TrimmingRadius=float("Inf"),
    GrowGrammars=np.array([]),
    ShrinkGrammars=np.array([]),
    GrammarOrder=["Grow", "Shrink"],
    #                 FastSolve = False,
    AvoidSolitary=False,
    EmbPointProb=1,
    PointWeights=None,
    AdjustElasticMatrix=None,
    AdjustElasticMatrix_Initial=None,
    DisplayWarnings=False,
    StoreGraphEvolution=False,
    GPU=False,
    FixNodesAtPoints=[],
    pseudotime=None,
    pseudotimeLambda=0.01,
    label=None,
    labelLambda=0.01,
    MaxNumberOfGraphCandidatesDict={
        "AddNode2Node": float("inf"),
        "BisectEdge": float("inf"),
        "RemoveNode": float("inf"),
        "ShrinkEdge": float("inf"),
    },
):
    """
    Core function to construct a principal elastic graph

    The core function that takes the n m-dimensional points and construct a principal elastic graph using the
    grammars provided.

    X numerical 2D matrix, the n-by-m matrix with the position of n m-dimensional points
    NumNodes integer, the number of nodes of the principal graph
    Lambda real, the lambda parameter used the compute the elastic energy
    Mu real, the lambda parameter used the compute the elastic energy
    CompileReport boolean, should a step-by-step report with various information on the
    contruction of the principal graph be compiled?
    ShowTimer boolean, should the time to construct the graph be computed and reported for each step?
    ComputeMSEP boolean, should MSEP be computed when building the report?
    GrowGrammars list of strings, the grammar to be used in the growth step
    ShrinkGrammars list of strings, the grammar to be used in the shrink step
    GrammarOrder vector of strings, the order of application of the grammars. Can be any combination of "Grow" and "Shrink"
    NodesPositions numerical 2D matrix, the k-by-m matrix with k m-dimensional positions of the nodes
    in the initial step
    ElasticMatrix numerical 2D matrix, the k-by-k elastic matrix
    n.cores either an integer (indicating the number of cores to used for the creation of a cluster) or
    cluster structure returned, e.g., by makeCluster. If a cluster structure is used, all the nodes must contains X
    (this is done using clusterExport)
    MinParOP integer, the minimum number of operations to use parallel computation
    MaxNumberOfIterations integer, maximum number of steps to embed the nodes in the data
    eps real, minimal relative change in the position of the nodes to stop embedment
    TrimmingRadius real, maximal distance of point from a node to affect its embedment
    NumEdges integer, the maximum nulber of edges
    Mode integer, the energy computation mode
    FastSolve boolean, should FastSolve be used when fitting the points to the data?
    ClusType string, the type of cluster to use. It can gbe either "Sock" or "Fork".
    Currently fork clustering only works in Linux
    AvoidSolitary boolean, should configurations with "solitary nodes", i.e., nodes without associted points be discarded?
    FinalEnergy string indicating the final elastic emergy associated with the configuration. Currently it can be "Base" or "Penalized"
    alpha positive numeric, the value of the alpha parameter of the penalized elastic energy
    beta positive numeric, the value of the beta parameter of the penalized elastic energy
    EmbPointProb numeric between 0 and 1. If less than 1 point will be sampled at each iteration.
    EmbPointProb indicates the probability of using each points. This is an *experimental* feature, which may
    helps speeding up the computation if a large number of points is present.
    MaxFailedOperations integer, the maximum allowed number of consecutive failed grammar operations,
    i.e. appplication of the single grammar operations, that did not produce any valid configuration
    MaxSteps integer, the maximum allowed number of steps of the algorithm. Each step is composed by the application of
    all the specified grammar operations
    GrammarOptimization boolean, should the grammar be used to optimize the graph? If True a number MaxSteps of operations will be applied.
    AdjustElasticMatrix a penalization function to adjust the elastic matrices after a configuration has been chosen (e.g., AdjustByConstant).
    If None (the default), no penalization will be used.
    AdjustVect boolean vector keeping track of the nodes for which the elasticity parameters have been adjusted.
    When True for a node its elasticity parameters will not be adjusted.
    gamma
    verbose
    AdjustElasticMatrix.Initial a penalization function to adjust the elastic matrices of the initial configuration (e.g., AdjustByConstant).
    If None (the default), no penalization will be used.

    Return
    -------
    a named list with a number of elements:
    \describe{
      \item{NodePositions}{A numeric matrix containing the positions of the nodes}
      \item{ElasticMatrix}{The elastic matrix of the graph}
      \item{ReportTable}{The report table for the graph construction}
      \item{FinalReport}{The report table associated with the final graph configuration}
      \item{Lambda}{The lambda parameter used during the graph construction}
      \item{Mu}{The mu parameter used during the graph construction}
      \item{FastSolve}{was FastSolve being used?}
    }



    @examples

    This is a low level function. See  \code{\link{computeElasticPrincipalCircle}},
    \code{\link{computeElasticPrincipalCurve}}, or \code{\link{computeElasticPrincipalTree}}


    """

    if PointWeights is None:
        PointWeights = np.ones((len(X), 1))

    if GrammarOptimization:
        print("Using grammar optimization")
        if np.isinf(MaxSteps):
            print(
                "When setting GrammarOptimization to TRUE, MaxSteps must be"
                " finite. Using MaxSteps = 1"
            )
            MaxSteps = 1

    if not isinstance(X, np.ndarray):
        raise TypeError("Please provide data matrix as an np array")

    if not CompileReport:
        verbose = False

    if "RemoveNode" in MaxNumberOfGraphCandidatesDict:
        if np.isfinite(MaxNumberOfGraphCandidatesDict["RemoveNode"]):
            raise ValueError(
                "MaxNumberOfGraphCandidates limitation not yet available for"
                " RemoveNode. Keep it to infinity"
            )

    if isinstance(ElasticMatrix, np.ndarray):
        if np.any(ElasticMatrix != ElasticMatrix.T):
            raise ValueError("Elastic matrix must be square and symmetric")

    if AdjustElasticMatrix_Initial is not None:
        ElasticMatrix, _ = AdjustElasticMatrix_Initial(
            ElasticMatrix, AdjustVect, verbose=True
        )

    ReportTable = []
    SquaredX = (X ** 2).sum(axis=1, keepdims=1)
    if GPU:
        Xcp = cupy.asarray(X)
        SquaredXcp = (Xcp ** 2).sum(axis=1, keepdims=1)
        InitNodePositions = PrimitiveElasticGraphEmbedment_cp(
            X=X,
            NodePositions=NodePositions,
            MaxNumberOfIterations=MaxNumberOfIterations,
            TrimmingRadius=TrimmingRadius,
            eps=eps,
            ElasticMatrix=ElasticMatrix,
            Mode=Mode,
            Xcp=Xcp,
            SquaredXcp=SquaredXcp,
            FixNodesAtPoints=FixNodesAtPoints,
            PointWeights=PointWeights,
        )[0]
    else:
        Xcp = None
        SquaredXcp = None
        InitNodePositions = PrimitiveElasticGraphEmbedment(
            X=X,
            NodePositions=NodePositions,
            MaxNumberOfIterations=MaxNumberOfIterations,
            TrimmingRadius=TrimmingRadius,
            eps=eps,
            ElasticMatrix=ElasticMatrix,
            Mode=Mode,
            FixNodesAtPoints=FixNodesAtPoints,
            PointWeights=PointWeights,
        )[0]

    UpdatedPG = dict(
        ElasticMatrix=ElasticMatrix,
        NodePositions=InitNodePositions,
        AdjustVect=AdjustVect,
    )

    #     if n_cores > 1:
    #         print('Copying data to shared memory for parallel processing...',end='')
    # #         ray.init(num_cpus=n_cores)
    # #         Xremote = ray.put(X)
    # #         SquaredXremote = ray.put(SquaredX)
    #         X_shape = X.shape
    #         SquaredX_shape = SquaredX.shape

    #         X_remote = multiprocessing.RawArray('d', X_shape[0] * X_shape[1])
    #         SquaredX_remote = multiprocessing.RawArray('d', SquaredX_shape[0] * SquaredX_shape[1])

    #         # Wrap remote objects as numpy arrays so we can easily manipulate their data.
    #         X_np = np.frombuffer(X_remote).reshape(X_shape)
    #         SquaredX_np = np.frombuffer(SquaredX_remote).reshape(SquaredX_shape)

    #         # Copy data to our shared array.
    #         np.copyto(X_np, X)
    #         np.copyto(SquaredX_np, SquaredX)
    #         multiproc_shared_variables = (X_remote,X_shape,SquaredX_remote,SquaredX_shape)
    #         # Initialize dictionary storing the variables passed from the init_worker.
    #         var_dict = {}

    #         print('Done')

    if verbose:
        print(
            "\nBARCODE\tENERGY\tNNODES\tNEDGES\tNRIBS\tNSTARS\tNRAYS\tNRAYS2\tMSE\tMSEP\tFVE\tFVEP\tUE\tUR\tURN\tURN2\tURSD\n"
        )

    # now we grow the graph up to NumNodes

    if (UpdatedPG["NodePositions"].shape[0] >= NumNodes) and not (GrammarOptimization):
        FinalReport = ReportOnPrimitiveGraphEmbedment(
            X=X,
            NodePositions=UpdatedPG["NodePositions"],
            ElasticMatrix=UpdatedPG["ElasticMatrix"],
            PartData=PartitionData(
                X=X,
                NodePositions=UpdatedPG["NodePositions"],
                MaxBlockSize=100000000,
                SquaredX=SquaredX,
                TrimmingRadius=TrimmingRadius,
            ),
            ComputeMSEP=ComputeMSEP,
            PointWeights=PointWeights,
        )

        return dict(
            NodePositions=UpdatedPG["NodePositions"],
            ElasticMatrix=UpdatedPG["ElasticMatrix"],
            ReportTable=FinalReport,
            FinalReport=FinalReport,
            Lambda=Lambda,
            Mu=Mu,
            Mode=Mode,
            MaxNumberOfIterations=MaxNumberOfIterations,
            eps=eps,
            times={},
            AllNodePositions={},
            AllElasticMatrices={},
            AllMergedElasticMatrices={},
            AllMergedNodePositions={},
        )

    FailedOperations = 0
    Steps = 0

    start = time.time()
    times = {}

    AllMergedElasticMatrices = {}
    AllMergedNodePositions = {}
    AllNodePositions = {}
    AllElasticMatrices = {}

    while (UpdatedPG["NodePositions"].shape[0] < NumNodes) or GrammarOptimization:
        nEdges = len(np.triu(UpdatedPG["ElasticMatrix"], 1).nonzero()[0])
        if (
            ((UpdatedPG["NodePositions"].shape[0]) >= NumNodes) or (nEdges >= NumEdges)
        ) and not GrammarOptimization:
            break

        OldPG = copy.deepcopy(UpdatedPG)

        for OpType in GrammarOrder:
            if OpType == "Grow" and len(GrowGrammars) > 0:

                for k in range(GrowGrammars.shape[0]):
                    if ShowTimer:
                        print("Growing")
                        t = time.time()

                    UpdatedPG = ApplyOptimalGraphGrammarOperation_v2(
                        X,
                        UpdatedPG["NodePositions"],
                        UpdatedPG["ElasticMatrix"],
                        GrowGrammars[k],
                        MaxBlockSize=MaxBlockSize,
                        AdjustVect=UpdatedPG["AdjustVect"],
                        SquaredX=SquaredX,
                        verbose=False,
                        MaxNumberOfIterations=MaxNumberOfIterations,
                        eps=eps,
                        TrimmingRadius=TrimmingRadius,
                        Mode=Mode,
                        FinalEnergy=FinalEnergy,
                        alpha=alpha,
                        beta=beta,
                        EmbPointProb=EmbPointProb,
                        PointWeights=PointWeights,
                        AvoidSolitary=AvoidSolitary,
                        AdjustElasticMatrix=AdjustElasticMatrix,
                        DisplayWarnings=DisplayWarnings,
                        n_cores=n_cores,
                        MinParOp=MinParOp,
                        Xcp=Xcp,
                        SquaredXcp=SquaredXcp,
                        FixNodesAtPoints=FixNodesAtPoints,
                        pseudotime=pseudotime,
                        pseudotimeLambda=pseudotimeLambda,
                        # label=label,
                        # labelLambda=labelLambda,
                        MaxNumberOfGraphCandidatesDict=MaxNumberOfGraphCandidatesDict,
                    )

                    if UpdatedPG == "failed operation":
                        print("failed operation")
                        FailedOperations += 1
                        UpdatedPG = copy.deepcopy(OldPG)
                        break
                    else:
                        FailedOperations = 0
                        if len(UpdatedPG["NodePositions"]) == 3:
                            # this is needed to erase the star elasticity coefficient which was initially assigned to both leaf nodes,
                            # one can erase this information after the number of nodes in the graph is > 2

                            inds = np.where(
                                np.sum(
                                    UpdatedPG["ElasticMatrix"]
                                    - np.diag(np.diag(UpdatedPG["ElasticMatrix"]))
                                    > 0,
                                    axis=0,
                                )
                                == 1
                            )

                            UpdatedPG["ElasticMatrix"][inds, inds] = 0

                    if ShowTimer:
                        elapsed = time.time() - t
                        print(np.round(elapsed, 4))

            if OpType == "Shrink" and len(ShrinkGrammars) > 0:
                for k in range(ShrinkGrammars.shape[0]):
                    if ShowTimer:
                        print("Shrinking")
                        t = time.time()
                    UpdatedPG = ApplyOptimalGraphGrammarOperation_v2(
                        X,
                        UpdatedPG["NodePositions"],
                        UpdatedPG["ElasticMatrix"],
                        ShrinkGrammars[k],
                        MaxBlockSize=MaxBlockSize,
                        AdjustVect=UpdatedPG["AdjustVect"],
                        SquaredX=SquaredX,
                        verbose=False,
                        MaxNumberOfIterations=MaxNumberOfIterations,
                        eps=eps,
                        TrimmingRadius=TrimmingRadius,
                        Mode=Mode,
                        FinalEnergy=FinalEnergy,
                        alpha=alpha,
                        beta=beta,
                        EmbPointProb=EmbPointProb,
                        PointWeights=PointWeights,
                        AvoidSolitary=AvoidSolitary,
                        AdjustElasticMatrix=AdjustElasticMatrix,
                        DisplayWarnings=DisplayWarnings,
                        n_cores=n_cores,
                        MinParOp=MinParOp,
                        Xcp=Xcp,
                        SquaredXcp=SquaredXcp,
                        FixNodesAtPoints=FixNodesAtPoints,
                        pseudotime=pseudotime,
                        pseudotimeLambda=pseudotimeLambda,
                        # label=label,
                        # labelLambda=labelLambda,
                        MaxNumberOfGraphCandidatesDict=MaxNumberOfGraphCandidatesDict,
                    )

                    if UpdatedPG == "failed operation":
                        print("failed operation")
                        FailedOperations += 1
                        UpdatedPG = copy.deepcopy(OldPG)
                        break
                    else:
                        FailedOperations = 0

                    if ShowTimer:
                        elapsed = time.time() - t
                        print(np.round(elapsed, 4))

        if CompileReport:
            if GPU:
                PartData = PartitionData_cp(
                    Xcp,
                    NodePositions=UpdatedPG["NodePositions"],
                    MaxBlockSize=1000000000,
                    SquaredXcp=SquaredXcp,
                    TrimmingRadius=TrimmingRadius,
                )
            else:
                PartData = PartitionData(
                    X,
                    NodePositions=UpdatedPG["NodePositions"],
                    MaxBlockSize=1000000000,
                    SquaredX=SquaredX,
                    TrimmingRadius=TrimmingRadius,
                )
            tReport = ReportOnPrimitiveGraphEmbedment(
                X=X,
                NodePositions=UpdatedPG["NodePositions"],
                ElasticMatrix=UpdatedPG["ElasticMatrix"],
                PartData=PartData,
                ComputeMSEP=ComputeMSEP,
            )

            FinalReport = copy.deepcopy(tReport)
            for k, v in tReport.items():
                if isnumeric(v):
                    tReport[k] = str(np.round(v, 4))
            ReportTable.append(tReport)

            if verbose:
                print("\t".join(tReport.values()))
        #                 print("\n")

        # Count the execution steps
        Steps += 1

        # If the number of execution steps is larger than MaxSteps stop the algorithm
        if Steps > MaxSteps or FailedOperations > MaxFailedOperations:
            break

        times[UpdatedPG["NodePositions"].shape[0]] = time.time() - start
        if StoreGraphEvolution:
            AllNodePositions[UpdatedPG["NodePositions"].shape[0]] = UpdatedPG[
                "NodePositions"
            ]
            AllElasticMatrices[UpdatedPG["NodePositions"].shape[0]] = UpdatedPG[
                "ElasticMatrix"
            ]
            AllMergedElasticMatrices[UpdatedPG["NodePositions"].shape[0]] = UpdatedPG[
                "StoreMergedElasticMatrix"
            ]
            AllMergedNodePositions[UpdatedPG["NodePositions"].shape[0]] = UpdatedPG[
                "StoreMergedNodePositions"
            ]

    if not verbose:
        if not CompileReport:
            if GPU:
                tReport = ReportOnPrimitiveGraphEmbedment(
                    X=X,
                    NodePositions=UpdatedPG["NodePositions"],
                    ElasticMatrix=UpdatedPG["ElasticMatrix"],
                    PartData=PartitionData_cp(
                        Xcp=Xcp,
                        NodePositions=UpdatedPG["NodePositions"],
                        SquaredXcp=SquaredXcp,
                        TrimmingRadius=TrimmingRadius,
                        MaxBlockSize=MaxBlockSize,
                    ),
                    ComputeMSEP=ComputeMSEP,
                )
            else:
                tReport = ReportOnPrimitiveGraphEmbedment(
                    X=X,
                    NodePositions=UpdatedPG["NodePositions"],
                    ElasticMatrix=UpdatedPG["ElasticMatrix"],
                    PartData=PartitionData(
                        X=X,
                        NodePositions=UpdatedPG["NodePositions"],
                        SquaredX=SquaredX,
                        TrimmingRadius=TrimmingRadius,
                        MaxBlockSize=MaxBlockSize,
                    ),
                    ComputeMSEP=ComputeMSEP,
                )

            FinalReport = copy.deepcopy(tReport)
            for k, v in tReport.items():
                if isnumeric(v):
                    tReport[k] = str(np.round(v, 4))

        else:
            tReport = ReportTable[-1]
        if verbose:
            print("\n")
            print(
                "BARCODE\tENERGY\tNNODES\tNEDGES\tNRIBS\tNSTARS\tNRAYS\tNRAYS2\tMSE\tMSEP\tFVE\tFVEP\tUE\tUR\tURN\tURN2\tURSD\n"
            )
            print("\t".join(tReport.values()))
            print("\n")

    if CompileReport:
        ReportTable = {k: [d[k] for d in ReportTable] for k in ReportTable[0]}

    #     if n_cores > 1:
    #         ray.shutdown()

    return dict(
        NodePositions=UpdatedPG["NodePositions"],
        ElasticMatrix=UpdatedPG["ElasticMatrix"],
        ReportTable=ReportTable,
        FinalReport=FinalReport,
        Lambda=Lambda,
        Mu=Mu,
        Mode=Mode,
        MaxNumberOfIterations=MaxNumberOfIterations,
        eps=eps,
        times=times,
        AllNodePositions=AllNodePositions,
        AllElasticMatrices=AllElasticMatrices,
        AllMergedElasticMatrices=AllMergedElasticMatrices,
        AllMergedNodePositions=AllMergedNodePositions,
    )


def computeElasticPrincipalGraph(
    Data,
    InitNodePositions,
    AdjustVect,
    InitEdges,
    NumNodes,
    NumEdges=float("inf"),
    ElasticMatrix=None,
    Lambda=0.01,
    Mu=0.1,
    MaxNumberOfIterations=100,
    eps=0.01,
    TrimmingRadius=float("inf"),
    Do_PCA=True,
    CenterData=True,
    ComputeMSEP=True,
    verbose=False,
    ShowTimer=False,
    ReduceDimension=None,
    # drawAccuracyComplexity = True,
    # drawPCAView = True,
    # drawEnergy = True,
    n_cores=1,
    # ClusType = "Sock",
    MinParOp=20,
    Mode=1,
    FinalEnergy="Base",
    alpha=0,
    beta=0,
    # gamma = 0,
    GrowGrammars=np.array([]),
    ShrinkGrammars=np.array([]),
    GrammarOptimization=False,
    MaxSteps=float("inf"),
    GrammarOrder=["Grow", "Shrink"],
    #                                 FastSolve = False,
    AvoidSolitary=False,
    EmbPointProb=1,
    PointWeights=None,
    AdjustElasticMatrix=None,
    AdjustElasticMatrix_Initial=None,
    Lambda_Initial=None,
    Mu_Initial=None,
    DisplayWarnings=False,
    StoreGraphEvolution=False,
    GPU=False,
    FixNodesAtPoints=[],
    pseudotime=None,
    pseudotimeLambda=0.01,
    label=None,
    labelLambda=0.01,
    MaxNumberOfGraphCandidatesDict={
        "AddNode2Node": float("inf"),
        "BisectEdge": float("inf"),
        "RemoveNode": float("inf"),
        "ShrinkEdge": float("inf"),
    },
):

    """
    Regularize data and construct a principal elastic graph

    This allow to perform basic data regularization before constructing a principla elastic graph.


    Data numerical 2D matrix, the n-by-m matrix with the position of n m-dimensional points
    NumNodes integer, the number of nodes of the principal graph
    Lambda real, the lambda parameter used the compute the elastic energy
    Mu real, the lambda parameter used the compute the elastic energy
    Do_PCA boolean, should data and initial node positions be PCA trnasformed?
    CenterData boolean, should data and initial node positions be centered?
    ComputeMSEP boolean, should MSEP be computed when building the report?
    ReduceDimension integer vector, vector of principal components to retain when performing
    dimensionality reduction. If None all the components will be used
    InitNodePositions numerical 2D matrix, the k-by-m matrix with k m-dimensional positions of the nodes
    in the initial step
    InitEdges numerical 2D matrix, the e-by-2 matrix with e end-points of the edges connecting the nodes
    ElasticMatrix numerical 2D matrix, the e-by-e matrix containing the elasticity parameters of the edges
    MaxNumberOfIterations integer, maximum number of steps to embed the nodes in the data
    eps real, minimal relative change in the position of the nodes to stop embedment
    TrimmingRadius real, maximal distance of point from a node to affect its embedment
    verbose boolean, should debugging information be reported?
    ShowTimer boolean, should the time to construct the graph be computed and reported for each step?
    n_cores either an integer (indicating the number of cores to used for the creation of a cluster) or
    cluster structure returned, e.g., by makeCluster. If a cluster structure is used, all the nodes must contains X
    (this is done using clusterExport)
    MinParOp integer, the minimum number of operations to use parallel computation
    GrowGrammars list of strings, the grammar to be used in the growth step
    ShrinkGrammars list of strings, the grammar to be used in the shrink step
    NumEdges integer, the maximum nulber of edges
    Mode integer, the energy computation mode
    AvoidSolitary boolean, should configurations with "solitary nodes", i.e., nodes without associated points be discarded?
    EmbPointProb numeric between 0 and 1. If less than 1 point will be sampled at each iteration.
    EmbPointProb indicates the probability of using each points. This is an *experimental* feature, which may
    helps speeding up the computation if a large number of points is present.
    FinalEnergy string indicating the final elastic emergy associated with the configuration. Currently it can be "Base" or "Penalized"
    alpha positive numeric, the value of the alpha parameter of the penalized elastic energy
    beta positive numeric, the value of the beta parameter of the penalized elastic energy
    ... optional parameter that will be passed to the AdjustHOS function
    AdjustVect boolean vector keeping track of the nodes for which the elasticity parameters have been adjusted.
    When True for a node its elasticity parameters will not be adjusted.
    AdjustElasticMatrix a penalization function to adjust the elastic matrices after a configuration has been chosen (e.g., AdjustByConstant).
    If None (the default), no penalization will be used.
    AdjustElasticMatrix.Initial a penalization function to adjust the elastic matrices of the initial configuration (e.g., AdjustByConstant).
    If None (the default), no penalization will be used.
    Lambda.Initial
    Mu.Initial

    Return
    -------
    a named list with a number of elements:
    \describe{
      \item{NodePositions}{A numeric matrix containing the positions of the nodes}
      \item{Edges}{A numeric matrix containing the pairs of nodes connected by edges}
      \item{ElasticMatrix}{The elastic matrix of the graph}
      \item{ReportTable}{The report table for the graph construction}
      \item{FinalReport}{The report table associated with the final graph configuration}
      \item{Lambda}{The lambda parameter used during the graph construction}
      \item{Mu}{The mu parameter used during the graph construction}
      \item{FastSolve}{was FastSolve being used?}
    }



    @examples

    This is a low level function. See  \code{\link{computeElasticPrincipalCircle}},
    \code{\link{computeElasticPrincipalCurve}}, or \code{\link{computeElasticPrincipalTree}}
    for examples


    """
    ST = datetime.datetime.today().strftime("%Y-%m-%d-%H:%M:%S")
    t = time.time()

    if ReduceDimension is None:
        ReduceDimension = np.array(range(np.min(Data.shape)))

    elif not Do_PCA:
        if verbose:
            print("Cannot reduce dimensionality witout doing PCA (parameter" " Do_PCA)")
            print("Dimensionality reduction will be ignored")
        ReduceDimension = np.array(range(np.min(Data.shape)))

    DataCenters = np.mean(Data, axis=0)
    if CenterData:
        Data = Data - DataCenters
        InitNodePositions = InitNodePositions - DataCenters

    if Do_PCA:
        if verbose:
            print("Performing PCA")

        if isinstance(ReduceDimension, float):
            if ReduceDimension < 1:
                if verbose:
                    print(
                        "Dimensionality reduction via ratio of explained"
                        " variance (full PCA will be computed)"
                    )
                vglobal, PCAData, explainedVariances = PCA(Data)
                ReduceDimension = range(
                    np.min(
                        np.where(
                            np.cumsum(explainedVariances) / explainedVariances.sum()
                            >= ReduceDimension
                        )
                    )
                    + 1
                )
                perc = (
                    explainedVariances[ReduceDimension].sum()
                    / explainedVariances.sum()
                    * 100
                )

                InitNodePositions = InitNodePositions.dot(vglobal)
            else:
                raise ValueError("if ReduceDimension is a single value it must be < 1")

        else:
            if max(ReduceDimension + 1) > min(Data.shape):
                if verbose:
                    print(
                        "Selected dimensions are outside of the available"
                        " range. ReduceDimension will be updated"
                    )
                ReduceDimension = [
                    i for i in ReduceDimension if i in range(min(Data.shape))
                ]
            if max(ReduceDimension + 1) > min(Data.shape) * 0.75:
                if verbose:
                    print("Using standard PCA")
                vglobal, PCAData, explainedVariances = PCA(Data)
                perc = (
                    explainedVariances[ReduceDimension].sum()
                    / explainedVariances.sum()
                    * 100
                )

                InitNodePositions = InitNodePositions.dot(vglobal)

            else:
                if verbose:
                    print("Centering data and using PCA with truncated SVD")
                if not CenterData:
                    # if data was not centered, center it (for SVD)
                    DataCenters = np.mean(Data, axis=0)
                    Data = Data - DataCenters
                    InitNodePositions = InitNodePositions - DataCenters
                PCAData, explainedVariances, U, S, Vt = TruncPCA(
                    Data,
                    algorithm="randomized",
                    n_components=max(ReduceDimension + 1),
                )
                ExpVariance = np.sum(np.var(Data, axis=0))
                perc = np.sum(explainedVariances) / ExpVariance * 100

                vglobal = Vt.T
                InitNodePositions = InitNodePositions.dot(vglobal)
        if verbose:
            print(len(ReduceDimension), "dimensions are being used")
            print(
                np.round(perc, 2),
                "% of the original variance has been retained",
            )

        X = PCAData[:, ReduceDimension]
        InitNodePositions = InitNodePositions[:, ReduceDimension]

    else:
        X = Data

    if Lambda_Initial is None:
        Lambda_Initial = Lambda

    if Mu_Initial is None:
        Mu_Initial = Mu

    if FixNodesAtPoints != []:
        nFixedNodes = len(FixNodesAtPoints)

        # fixed datapoints
        flat_FixNodesAtPoints = [
            item for sublist in FixNodesAtPoints for item in sublist
        ]
        if len(set(flat_FixNodesAtPoints)) != len(flat_FixNodesAtPoints):
            raise ValueError("FixNodesAtPoints lists contain duplicate points")

        # init fixed nodes
        FixedNodePositions = np.array(
            [X[inds].mean(axis=0) for inds in FixNodesAtPoints]
        )

        # init edges at first indices (with closest NodePosition)
        closest_node = np.zeros(nFixedNodes, dtype=int)
        for i, fn in enumerate(FixedNodePositions):
            closest_node[i] = np.argmin(
                [np.linalg.norm(fn - n) for n in InitNodePositions]
            )

        if ElasticMatrix is not None:
            InitEdges, _, _ = DecodeElasticMatrix(ElasticMatrix)
        AddEdges = np.vstack((np.arange(nFixedNodes), closest_node + nFixedNodes)).T

        # concatenate
        InitNodePositions = np.concatenate((FixedNodePositions, InitNodePositions))
        InitEdges = np.concatenate((AddEdges, InitEdges + nFixedNodes))
        ElasticMatrix = MakeUniformElasticMatrix(
            Edges=InitEdges, Lambda=Lambda_Initial, Mu=Mu_Initial
        )

    if ElasticMatrix is None:
        InitElasticMatrix = MakeUniformElasticMatrix(
            Edges=InitEdges, Lambda=Lambda_Initial, Mu=Mu_Initial
        )
    else:
        if verbose:
            print(
                "The elastic matrix is being used. Edge configuration will be"
                " ignored"
            )
        InitElasticMatrix = ElasticMatrix

    if (
        InitElasticMatrix.shape[0] != InitNodePositions.shape[0]
        or InitElasticMatrix.shape[1] != InitNodePositions.shape[0]
    ):
        raise ValueError(
            "Elastic matrix incompatible with the node number. Impossible to"
            " proceed."
        )

    # Computing the graph
    if verbose:
        print(
            "Computing EPG with ",
            NumNodes,
            " nodes on ",
            Data.shape[0],
            " points and ",
            Data.shape[1],
            " dimensions",
        )

    ElData = ElPrincGraph(
        X=X,
        NumNodes=NumNodes,
        NumEdges=NumEdges,
        Lambda=Lambda,
        Mu=Mu,
        MaxNumberOfIterations=MaxNumberOfIterations,
        eps=eps,
        TrimmingRadius=TrimmingRadius,
        NodePositions=InitNodePositions,
        ElasticMatrix=InitElasticMatrix,
        AdjustVect=AdjustVect,
        CompileReport=True,
        ShowTimer=ShowTimer,
        FinalEnergy=FinalEnergy,
        alpha=alpha,
        beta=beta,
        Mode=Mode,
        GrowGrammars=GrowGrammars,
        ShrinkGrammars=ShrinkGrammars,
        GrammarOptimization=GrammarOptimization,
        MaxSteps=MaxSteps,
        GrammarOrder=GrammarOrder,
        ComputeMSEP=ComputeMSEP,
        verbose=verbose,
        AvoidSolitary=AvoidSolitary,
        EmbPointProb=EmbPointProb,
        PointWeights=PointWeights,
        AdjustElasticMatrix=AdjustElasticMatrix,
        AdjustElasticMatrix_Initial=AdjustElasticMatrix_Initial,
        DisplayWarnings=DisplayWarnings,
        n_cores=n_cores,
        MinParOp=MinParOp,
        StoreGraphEvolution=StoreGraphEvolution,
        GPU=GPU,
        FixNodesAtPoints=FixNodesAtPoints,
        pseudotime=pseudotime,
        pseudotimeLambda=pseudotimeLambda,
        label=label,
        labelLambda=labelLambda,
        MaxNumberOfGraphCandidatesDict=MaxNumberOfGraphCandidatesDict,
    )

    Edges = DecodeElasticMatrix(ElData["ElasticMatrix"])
    NodePositions = ElData["NodePositions"]
    if "AllNodePositions" not in ElData.keys():
        AllNodePositions = {}
        AllMergedNodePositions = {}
    else:
        AllNodePositions = ElData["AllNodePositions"]
        AllMergedNodePositions = ElData["AllMergedNodePositions"]

    # if drawEnergy and ElData['ReportTable'] is not None:
    #    print('MSDEnergyPlot not yet implemented')
    #     plotMSDEnergyPlot(ReportTable = ElData['ReportTable'])

    # if drawAccuracyComplexity and ElData['ReportTable'] is not None:
    #    print('accuracyComplexityPlot not yet implemented')
    #     accuracyComplexityPlot(ReportTable = ElData['ReportTable'])

    if Do_PCA:
        NodePositions = NodePositions.dot(vglobal[:, ReduceDimension].T)
        for (k, nodep), (_, allnodep) in zip(
            AllNodePositions.items(), AllMergedNodePositions.items()
        ):
            AllNodePositions[k] = nodep.dot(vglobal[:, ReduceDimension].T)
            if AllMergedNodePositions[k] is not None:
                AllMergedNodePositions[k] = allnodep.dot(vglobal[:, ReduceDimension].T)

    EndTimer = time.time() - t
    if verbose:
        print(np.round(EndTimer, 4), " seconds elapsed")

    FinalPG = dict(
        NodePositions=NodePositions,
        Edges=Edges,
        ReportTable=ElData["ReportTable"],
        FinalReport=ElData["FinalReport"],
        ElasticMatrix=ElData["ElasticMatrix"],
        Lambda=ElData["Lambda"],
        Mu=ElData["Mu"],
        TrimmingRadius=TrimmingRadius,
        Mode=ElData["Mode"],
        MaxNumberOfIterations=ElData["MaxNumberOfIterations"],
        eps=ElData["eps"],
        Date=ST,
        TicToc=EndTimer,
        times=ElData["times"],
        AllNodePositions=AllNodePositions,
        AllElasticMatrices=ElData["AllElasticMatrices"],
        AllMergedElasticMatrices=ElData["AllMergedElasticMatrices"],
        AllMergedNodePositions=AllMergedNodePositions,
    )

    # if drawPCAView:
    #    print(PlotPG(Data, FinalPG))

    if Do_PCA or CenterData:
        FinalPG["NodePositions"] = NodePositions + DataCenters
        for (k, nodep), (_, allnodep) in zip(
            FinalPG["AllNodePositions"].items(),
            FinalPG["AllMergedNodePositions"].items(),
        ):
            FinalPG["AllNodePositions"][k] = nodep + DataCenters
            if FinalPG["AllMergedNodePositions"][k] is not None:
                FinalPG["AllMergedNodePositions"][k] = allnodep + DataCenters
    return FinalPG
